from mindsdb.interfaces.storage import db
from mindsdb.interfaces.query_context.context_controller import query_context_controller
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError


class ViewController:
    def add(self, name, query, project_name):
        from mindsdb.interfaces.database.database import DatabaseController

        database_controller = DatabaseController()
        project_databases_dict = database_controller.get_dict(filter_type='project')

        if project_name not in project_databases_dict:
            raise EntityNotExistsError('Can not find project', project_name)

        project_id = project_databases_dict[project_name]['id']
        view_record = (
            db.session.query(db.View.id)
            .filter_by(
                name=name,
                company_id=ctx.company_id,
                project_id=project_id
            ).first()
        )
        if view_record is not None:
            raise EntityExistsError('View already exists', name)

        view_record = db.View(
            name=name,
            company_id=ctx.company_id,
            query=query,
            project_id=project_id
        )
        db.session.add(view_record)
        db.session.commit()

    def update(self, name, query, project_name):
        project_record = db.session.query(db.Project).filter_by(
            name=project_name,
            company_id=ctx.company_id,
            deleted_at=None
        ).first()
        rec = db.session.query(db.View).filter(
            db.View.name == name,
            db.View.company_id == ctx.company_id,
            db.View.project_id == project_record.id
        ).first()
        if rec is None:
            raise EntityNotExistsError('View not found', name)
        rec.query = query
        db.session.commit()

    def delete(self, name, project_name):
        project_record = db.session.query(db.Project).filter_by(
            name=project_name,
            company_id=ctx.company_id,
            deleted_at=None
        ).first()
        rec = db.session.query(db.View).filter(
            db.View.name == name,
            db.View.company_id == ctx.company_id,
            db.View.project_id == project_record.id
        ).first()
        if rec is None:
            raise EntityNotExistsError('View not found', name)
        db.session.delete(rec)
        db.session.commit()

        query_context_controller.drop_query_context('view', rec.id)

    def list(self, project_name):
        query = db.session.query(db.Project).filter_by(
            company_id=ctx.company_id,
            deleted_at=None
        )
        if project_name is not None:
            query = query.filter_by(name=project_name)

        project_names = {
            i.id: i.name
            for i in query
        }

        query = db.session.query(db.View).filter(
            db.View.company_id == ctx.company_id,
            db.View.project_id.in_(list(project_names.keys()))
        )

        data = []

        for record in query:

            data.append({
                'id': record.id,
                'name': record.name,
                'project': project_names[record.project_id],
                'query': record.query,
            })

        return data

    def _get_view_record_data(self, record):
        return {
            'id': record.id,
            'name': record.name,
            'query': record.query
        }

    def get(self, id=None, name=None, project_name=None):
        project_record = db.session.query(db.Project).filter_by(
            name=project_name,
            company_id=ctx.company_id,
            deleted_at=None
        ).first()
        if id is not None:
            records = db.session.query(db.View).filter_by(
                id=id,
                project_id=project_record.id,
                company_id=ctx.company_id
            ).all()
        elif name is not None:
            records = db.session.query(db.View).filter_by(
                name=name,
                project_id=project_record.id,
                company_id=ctx.company_id
            ).all()
        if len(records) == 0:
            if name is None:
                name = f'id={id}'
            raise EntityNotExistsError("Can't find view", f'{project_name}.{name}')
        elif len(records) > 1:
            raise Exception(f"There are multiple views with name/id: {name}/{id}")
        record = records[0]
        return self._get_view_record_data(record)



from mindsdb.integrations.handlers.postgres_handler import Handler as PostgresHandler


class OrioleDBHandler(PostgresHandler):
    """
    This handler handles connection and execution of the OrioleDB statements.
    """
    name = 'orioledb'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



import os
import sys
import subprocess


def install_dependencies(dependencies):
    outs = b''
    errs = b''
    result = {
        'success': False,
        'error_message': None
    }

    # get the path to this script
    script_path = os.path.dirname(os.path.realpath(__file__))
    split_dependencies = []
    for dependency in dependencies:
        # ignore standalone comments
        if dependency.startswith('#'):
            continue

        # remove inline comments
        if '#' in dependency:
            dependency = dependency.split('#')[0].strip()

        # check if the dependency is a path to a requirements file
        if dependency.startswith('-r'):
            # split the string into '-r' and the path
            r_flag, req_path = dependency.split(' ')
            # create the absolute path to the requirements file
            abs_req_path = os.path.abspath(os.path.join(script_path, req_path.replace('mindsdb/integrations', '..')))

            # check if the file exists
            if os.path.exists(abs_req_path):
                split_dependencies.append(r_flag)
                split_dependencies.append(abs_req_path)
            else:
                raise FileNotFoundError(f"Requirements file not found: {req_path}")

        else:
            split_dependencies.append(dependency)

    try:
        sp = subprocess.Popen(
            [sys.executable, '-m', 'pip', 'install', *split_dependencies],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        code = sp.wait()
        outs, errs = sp.communicate(timeout=1)
    except Exception as e:
        result['error_message'] = str(e)

    if code != 0:
        output = ''
        if isinstance(outs, bytes) and len(outs) > 0:
            output = output + 'Output: ' + outs.decode()
        if isinstance(errs, bytes) and len(errs) > 0:
            if len(output) > 0:
                output = output + '\n'
            output = output + 'Errors: ' + errs.decode()
        result['error_message'] = output
    else:
        result['success'] = True

    return result



from mindsdb.utilities import log
from google.cloud import aiplatform
import pandas as pd

from mindsdb.integrations.utilities.handlers.auth_utilities import GoogleServiceAccountOAuth2Manager

logger = log.getLogger(__name__)


class VertexClient:
    """A class to interact with Vertex AI"""

    def __init__(self, args_json, credentials_url=None, credentials_file=None, credentials_json=None):
        google_sa_oauth2_manager = GoogleServiceAccountOAuth2Manager(
            credentials_url=credentials_url,
            credentials_file=credentials_file,
            credentials_json=credentials_json,
        )
        credentials = google_sa_oauth2_manager.get_oauth2_credentials()

        aiplatform.init(
            credentials=credentials,
            project=args_json["project_id"],
            location=args_json["location"],
            staging_bucket=args_json["staging_bucket"],
            # the name of the experiment to use to track
            # logged metrics and parameters
            experiment=args_json["experiment"],
            # description of the experiment above
            experiment_description=args_json["experiment_description"],
        )

    def print_datasets(self):
        """Print all datasets and dataset ids in the project"""
        for dataset in aiplatform.TabularDataset.list():
            logger.info(f"Dataset display name: {dataset.display_name}, ID: {dataset.name}")

    def print_models(self):
        """Print all model names and model ids in the project"""
        for model in aiplatform.Model.list():
            logger.info(f"Model display name: {model.display_name}, ID: {model.name}")

    def print_endpoints(self):
        """Print all endpoints and endpoint ids in the project"""
        for endpoint in aiplatform.Endpoint.list():
            logger.info(f"Endpoint display name: {endpoint.display_name}, ID: {endpoint.name}")

    def get_model_by_display_name(self, display_name):
        """Get a model by its display name"""
        try:
            return aiplatform.Model.list(filter=f"display_name={display_name}")[0]
        except IndexError:
            logger.info(f"Model with display name {display_name} not found")

    def get_endpoint_by_display_name(self, display_name):
        """Get an endpoint by its display name"""
        try:
            return aiplatform.Endpoint.list(filter=f"display_name={display_name}")[0]
        except IndexError:
            logger.info(f"Endpoint with display name {display_name} not found")

    def get_model_by_id(self, model_id):
        """Get a model by its ID"""
        try:
            return aiplatform.Model(model_name=model_id)
        except IndexError:
            logger.info(f"Model with ID {model_id} not found")

    def deploy_model(self, model):
        """Deploy a model to an endpoint - long runtime"""
        endpoint = model.deploy()
        return endpoint

    def predict_from_df(self, endpoint_display_name, df, custom_model=False):
        """Make a prediction from a Pandas dataframe"""
        endpoint = self.get_endpoint_by_display_name(endpoint_display_name)
        if custom_model:
            records = df.values.tolist()
        else:
            records = df.astype(str).to_dict(orient="records")  # list of dictionaries
        prediction = endpoint.predict(instances=records)
        return prediction

    def predict_from_csv(self, endpoint_display_name, csv_to_predict):
        """Make a prediction from a CSV file"""
        df = pd.read_csv(csv_to_predict)
        return self.predict_from_df(endpoint_display_name, df)

    def predict_from_dict(self, endpoint_display_name, data):

        # convert to list of dictionaries
        instances = [dict(zip(data.keys(), values)) for values in zip(*data.values())]
        endpoint = self.get_endpoint_by_display_name(endpoint_display_name)
        prediction = endpoint.predict(instances=instances)
        return prediction



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database file to read and write from. The special value :memory: (default) can be used to create an in-memory database.',
    },
    read_only={
        'type': ARG_TYPE.BOOL,
        'description': 'A flag that specifies if the connection should be made in read-only mode.',
    },
)

connection_args_example = OrderedDict(database='db.duckdb', read_only=True)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host = {
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Altibase server. NOTE: use \'127.0.0.1\' instead of \'localhost\' to connect to local server.'
    },
    port = {
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the Altibase server. Must be an integer.'
    },
    user = {
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Altibase server.'
    },
    password = {
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Altibase server.',
        'secret': True
    },
    database = {
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the Altibase server.'
    },
    jdbcClass = {
        'type': ARG_TYPE.STR,
        'description': 'The driver class of the Altibase JDBC driver'
    },
    jdbcJarLocation = {
        'type': ARG_TYPE.PATH,
        'description': 'The location of the Altibase JDBC driver jar file'
    },
    dsn = {
        'type': ARG_TYPE.STR,
        'description': 'Datasource name of the Altibase server. NOTE: use dsn if you want to use an ODBC connection.'
    }
)

connection_args_example = OrderedDict(
    host = '127.0.0.1',
    port = 20300,
    user = 'sys',
    password = 'manager',
    database = 'mydb',
    jdbcClass='Altibase.jdbc.driver.AltibaseDriver',
    jdbcJarLocation='/data/altibase_home/lib/Altibase.jar'
)



from typing import List
from pydantic_settings import BaseSettings


class MSTeamsHandlerConfig(BaseSettings):
    """
    Settings for the Microsoft Teams handler.

    Attributes
    ----------
    DEFAULT_SCOPES: List
        Default scopes for querying the Microsoft Graph API.

    CHATS_TABLE_COLUMNS: List
        Columns for the chats table. 

    CHAT_MESSAGES_TABLE_COLUMNS: List
        Columns for the chat messages table.

    CHANNELS_TABLE_COLUMNS: List
        Columns for the channels table.

    CHANNEL_MESSAGES_TABLE_COLUMNS: List
        Columns for the channel messages table.
    
    TEST_CHAT_DATA: Dict
        Test data for the chats table.

    TEST_CHATS_DATA: List[Dict]
        Test data for the chats table.

    TEST_CHAT_MESSAGE_DATA: Dict
        Test data for the chat messages table.

    TEST_CHAT_MESSAGES_DATA: List[Dict]
        Test data for the chat messages table.

    TEST_CHANNEL_DATA: Dict
        Test data for the channels table.

    TEST_CHANNELS_DATA: List[Dict]
        Test data for the channels table.

    TEST_CHANNEL_MESSAGE_DATA: Dict
        Test data for the channel messages table.  

    TEST_CHANNEL_MESSAGES_DATA: List[Dict]
        Test data for the channel messages table.
    
    TEST_GROUP_DATA: dict

    TEST_CHANNEL_ID_DATA: dict
    """

    DEFAULT_SCOPES: List = [
        'https://graph.microsoft.com/User.Read',
        'https://graph.microsoft.com/Group.Read.All',
        'https://graph.microsoft.com/ChannelMessage.Send',
        'https://graph.microsoft.com/Chat.Read',
        'https://graph.microsoft.com/ChatMessage.Send',
    ]

    CHATS_TABLE_COLUMNS: List = [
        "id",
        "topic",
        "createdDateTime",
        "lastUpdatedDateTime",
        "chatType",
        "webUrl",
        "tenantId",
        "onlineMeetingInfo",
        "viewpoint_isHidden",
        "viewpoint_lastMessageReadDateTime",
        "lastMessagePreview_id",
        "lastMessagePreview_createdDateTime",
    ]

    CHAT_MESSAGES_TABLE_COLUMNS: List = [
        "id",
        "replyToId",
        "etag",
        "messageType",
        "createdDateTime",
        "lastModifiedDateTime",
        "lastEditedDateTime",
        "deletedDateTime",
        "subject",
        "summary",
        "chatId",
        "importance",
        "locale",
        "webUrl",
        "channelIdentity",
        "policyViolation",
        "attachments",
        "mentions",
        "reactions",
        "from_application",
        "from_device",
        "from_user_@odata.type",
        "from_user_id",
        "from_user_displayName",
        "from_user_userIdentityType",
        "from_user_tenantId",
        "body_contentType",
        "body_content"
    ]

    CHANNELS_TABLE_COLUMNS: List = [
        "id",
        "createdDateTime",
        "displayName",
        "description",
        "isFavoriteByDefault",
        "email",
        "tenantId",
        "webUrl",
        "membershipType",
        "teamId",
    ]

    CHANNEL_MESSAGES_TABLE_COLUMNS: List = [
        "id",
        "replyToId",
        "etag",
        "messageType",
        "createdDateTime",
        "lastModifiedDateTime",
        "lastEditedDateTime",
        "deletedDateTime",
        "subject",
        "summary",
        "chatId",
        "importance",
        "locale",
        "webUrl",
        "policyViolation",
        "attachments",
        "mentions",
        "reactions",
        "from_application",
        "from_device",
        "from_user_@odata.type",
        "from_user_id",
        "from_user_displayName",
        "from_user_userIdentityType",
        "from_user_tenantId",
        "body_contentType",
        "body_content",
        "channelIdentity_teamId",
        "channelIdentity_channelId",
        "eventDetail_@odata.type",
        "eventDetail_visibleHistoryStartDateTime",
        "eventDetail_members",
        "eventDetail_initiator_device",
        "eventDetail_initiator_application_@odata.type",
        "eventDetail_initiator_application_id",
        "eventDetail_initiator_application_displayName",
        "eventDetail_initiator_application_applicationIdentityType",
        "eventDetail_channelId",
        "eventDetail_channelDisplayName",
        "eventDetail_initiator_user_@odata.type",
        "eventDetail_initiator_user_id",
        "eventDetail_initiator_user_displayName",
        "eventDetail_initiator_user_userIdentityType",
        "eventDetail_initiator_user_tenantId"
    ]

    TEST_CHAT_DATA: dict = {
        '@odata.context': 'test_context',
        'id': 'test_id',
        'topic': None,
        'createdDateTime': '2023-11-20T10:25:19.553Z',
        'lastUpdatedDateTime': '2023-11-20T10:25:19.553Z',
        'chatType': 'oneOnOne',
        'webUrl': 'https://teams.test',
        'tenantId': 'test_tenant_id',
        'onlineMeetingInfo': None,
        'viewpoint': {
            'isHidden': False,
            'lastMessageReadDateTime': '2023-12-08T17:09:34.214Z'
        },
        'lastMessagePreview@odata.context': 'https://graph.test',
        'lastMessagePreview': {
            'id': '1702055374214',
            'createdDateTime': '2023-12-08T17:09:34.214Z',
            'isDeleted': False,
            'messageType': 'message',
            'eventDetail': None,
            'body': {
                'contentType': 'text',
                'content': '\n\nTest message.'
            },
            'from': {
                'application': None,
                'device': None,
                'user': {}
            }
        }
    }

    TEST_CHATS_DATA: List[dict] = [TEST_CHAT_DATA]

    TEST_CHAT_MESSAGE_DATA: dict = {
        '@odata.context': 'test_context', 
        'id': 'test_id', 
        'replyToId': None, 
        'etag': 'test_etag', 
        'messageType': 'message', 
        'createdDateTime': '2023-12-08T17:09:22.241Z', 
        'lastModifiedDateTime': '2023-12-08T17:09:22.241Z', 
        'lastEditedDateTime': None, 
        'deletedDateTime': None, 
        'subject': None, 
        'summary': None, 
        'chatId': 'test_chat_id', 
        'importance': 'normal', 
        'locale': 'en-us',
        'webUrl': None,
        'channelIdentity': None,
        'policyViolation': None,
        'attachments': [],
        'mentions': [],
        'reactions': [],
        'from': {
            'application': None,
            'device': None,
            'user': {
                '@odata.type': 'test_type',
                'id': 'test_user_id',
                'displayName': 'test_user_display_name',
                'userIdentityType': 'aadUser',
                'tenantId': 'test_tenant_id'
            }
        },
        'body': {
            'contentType': 'text',
            'content': '\n\nTest message.'
        },
        'eventDetail': {
            '@odata.type': 'test_type',
            'visibleHistoryStartDateTime': '2023-12-08T17:09:22.241Z',
            'members': [],
            'initiator': {
                'device': None,
                'application': None,
                'user': {
                    '@odata.type': 'test_type',
                    'id': 'test_user_id',
                    'displayName': 'test_user_display_name',
                    'userIdentityType': 'aadUser',
                    'tenantId': 'test_tenant_id'
                }
            },
        }
    }

    TEST_CHAT_MESSAGES_DATA: List[dict] = [TEST_CHAT_MESSAGE_DATA]

    TEST_CHANNEL_DATA: dict = {
        '@odata.context': 'test_context', 
        'id': 'test_id', 
        'createdDateTime': '2023-11-17T22:54:33.055Z', 
        'displayName': 'test_display_name', 
        'description': None, 
        'isFavoriteByDefault': None, 
        'email': 'test@test.com', 
        'tenantId': 'test_tenant_id', 
        'webUrl': 'https://teams.test', 
        'membershipType': 'standard', 
        'teamId': 'test_team_id'
    }

    TEST_CHANNELS_DATA: List[dict] = [TEST_CHANNEL_DATA]

    TEST_CHANNEL_MESSAGE_DATA: dict = {
        '@odata.context': 'test_context', 
        'id': 'test_id', 
        'replyToId': None, 
        'etag': 'test_etag', 
        'messageType': 'message', 
        'createdDateTime': '2023-11-30T16:52:50.18Z', 
        'lastModifiedDateTime': '2023-11-30T16:52:50.18Z', 
        'lastEditedDateTime': None, 
        'deletedDateTime': None, 
        'subject': 'Test Subject', 
        'summary': None, 
        'chatId': None, 
        'importance': 
        'normal', 
        'locale': 'en-us',
        'webUrl': 'https://teams.test',
        'policyViolation': None,
        'attachments': [],
        'mentions': [],
        'reactions': [],
        'from': {
            'application': None,
            'device': None,
            'user': {
                '@odata.type': 'test_type',
                'id': 'test_user_id',
                'displayName': 'test_user_display_name',
                'userIdentityType': 'aadUser',
                'tenantId': 'test_tenant_id'
            }
        },
        'body': {
            'contentType': 'text',
            'content': '\n\nTest message.'
        },
        'channelIdentity': {
            'teamId': 'test_team_id',
            'channelId': 'test_channel_id'
        },
        'eventDetail': {
            '@odata.type': 'test_type',
            'visibleHistoryStartDateTime': '2023-11-30T16:52:50.18Z',
            'members': [],
            'initiator': {
                'device': None,
                'application': {
                    '@odata.type': 'test_type',
                    'id': 'test_app_id',
                    'displayName': 'test_app_display_name',
                    'applicationIdentityType': 'bot'
                },
                'user': {
                    '@odata.type': 'test_type',
                    'id': 'test_user_id',
                    'displayName': 'test_user_display_name',
                    'userIdentityType': 'aadUser',
                    'tenantId': 'test_tenant_id'
                }
            },
            'channelId': 'test_channel_id',
            'channelDisplayName': 'test_channel_display_name'
        }
    }

    TEST_CHANNEL_MESSAGES_DATA: List[dict] = [TEST_CHANNEL_MESSAGE_DATA]

    TEST_GROUP_DATA: dict = {
        '@odata.context': 'test_context', 
        'value': [
            {
                'id': 'test_team_id',
                'resourceProvisioningOptions': ['Team'],
            }
        ]
    }

    TEST_CHANNEL_ID_DATA: dict = {
        '@odata.context': 'test_context', 
        '@odata.count': 1,
        'value': [
            {
                'id': 'test_channel_id',
            }
        ]
    }

ms_teams_handler_config = MSTeamsHandlerConfig()


import enum
from typing import List, Optional

from mindsdb_sql.parser.ast import Select, BinaryOperation, Identifier, Constant, Star

from mindsdb.integrations.libs.vectordatabase_handler import TableField
from mindsdb.interfaces.storage import db
from .sql_agent import SQLAgent

_DEFAULT_TOP_K_SIMILARITY_SEARCH = 5
_DEFAULT_SQL_LLM_MODEL = 'gpt-3.5-turbo'


class SkillType(enum.Enum):
    TEXT2SQL_LEGACY = 'text2sql'
    TEXT2SQL = 'sql'
    KNOWLEDGE_BASE = 'knowledge_base'
    RETRIEVAL = 'retrieval'


class SkillToolController:
    def __init__(self):
        self.command_executor = None

    def get_command_executor(self):
        if self.command_executor is None:
            from mindsdb.api.executor.controllers import SessionController
            from mindsdb.api.executor.command_executor import ExecuteCommands

            sql_session = SessionController()
            sql_session.database = 'mindsdb'

            self.command_executor = ExecuteCommands(sql_session)
        return self.command_executor

    def get_sql_agent(
        self,
        database: str,
        include_tables: Optional[List[str]] = None,
        ignore_tables: Optional[List[str]] = None,
        sample_rows_in_table_info: int = 3,
    ):
        return SQLAgent(
            self.get_command_executor(),
            database,
            include_tables,
            ignore_tables,
            sample_rows_in_table_info,
        )

    def _make_text_to_sql_tools(self, skill: db.Skills, llm) -> dict:
        '''
           Uses SQLAgent to execute tool
        '''
        # To prevent dependency on Langchain unless an actual tool uses it.
        try:
            from mindsdb.integrations.handlers.langchain_handler.mindsdb_database_agent import MindsDBSQL
            from langchain_community.agent_toolkits.sql.toolkit import SQLDatabaseToolkit
            from langchain_community.tools.sql_database.tool import QuerySQLDataBaseTool
        except ImportError:
            raise ImportError('To use the text-to-SQL skill, please install langchain with `pip install mindsdb[langchain]`')
        database = skill.params['database']
        tables = skill.params['tables']
        tables_to_include = [f'{database}.{table}' for table in tables]
        db = MindsDBSQL(
            engine=self.get_command_executor(),
            database=database,
            metadata=self.get_command_executor().session.integration_controller,
            include_tables=tables_to_include
        )
        # Users probably don't need to configure this for now.
        sql_database_tools = SQLDatabaseToolkit(db=db, llm=llm).get_tools()
        description = skill.params.get('description', '')
        tables_list = ','.join([f'{database}.{table}' for table in tables])
        for i, tool in enumerate(sql_database_tools):
            if isinstance(tool, QuerySQLDataBaseTool):
                # Add our own custom description so our agent knows when to query this table.
                tool.description = (
                    f'Use this tool if you need data about {description}. '
                    'Use the conversation context to decide which table to query. '
                    f'These are the available tables: {tables_list}.\n'
                    f'ALWAYS consider these special cases:\n'
                    f'- Not all SQL functions are supported. Do NOT use the following functions: INTERVAL. \n'
                    f'- For TIMESTAMP type columns, make sure you include the time portion in your query (e.g. WHERE date_column = "2020-01-01 12:00:00")'
                    f'Here are the rest of the instructions:\n'
                    f'{tool.description}'
                )
                sql_database_tools[i] = tool
        return sql_database_tools

    def _make_retrieval_tools(self, skill: db.Skills) -> dict:
        """
        creates advanced retrieval tool i.e. RAG
        """
        params = skill.params
        return dict(
            name=params.get('name', skill.name),
            source=params.get('source', None),
            config=params.get('config', {}),
            description=f'You must use this tool to get more context or information '
                        f'to answer a question about {params["description"]}. '
                        f'The input should be the exact question the user is asking.',
            type=skill.type
        )

    def _get_rag_query_function(self, skill: db.Skills):

        session_controller = self.get_command_executor().session

        def _answer_question(question: str) -> str:
            knowledge_base_name = skill.params['source']

            # make select in KB table
            query = Select(
                targets=[Star()],
                where=BinaryOperation(op='=', args=[
                    Identifier(TableField.CONTENT.value), Constant(question)
                ]),
                limit=Constant(_DEFAULT_TOP_K_SIMILARITY_SEARCH),
            )
            kb_table = session_controller.kb_controller.get_table(knowledge_base_name, skill.project_id)

            res = kb_table.select_query(query)
            return '\n'.join(res.content)

        return _answer_question

    def _make_knowledge_base_tools(self, skill: db.Skills) -> dict:
        # To prevent dependency on Langchain unless an actual tool uses it.
        description = skill.params.get('description', '')

        return dict(
            name='Knowledge Base Retrieval',
            func=self._get_rag_query_function(skill),
            description=f'Use this tool to get more context or information to answer a question about {description}. The input should be the exact question the user is asking.',
            type=skill.type
        )

    def get_tools_from_skill(self, skill: db.Skills, llm) -> dict:
        """
            Creates function for skill and metadata (name, description)
        Args:
            skill (Skills): Skill to make a tool from

        Returns:
            dict with keys: name, description, func
        """

        try:
            skill_type = SkillType(skill.type)
        except ValueError:
            raise NotImplementedError(
                f'skill of type {skill.type} is not supported as a tool, supported types are: {list(SkillType._member_names_)}')

        if skill_type == SkillType.TEXT2SQL or skill_type == SkillType.TEXT2SQL_LEGACY:
            return self._make_text_to_sql_tools(skill, llm)
        if skill_type == SkillType.KNOWLEDGE_BASE:
            return [self._make_knowledge_base_tools(skill)]
        if skill_type == SkillType.RETRIEVAL:
            return [self._make_retrieval_tools(skill)]


skill_tool = SkillToolController()



"""predictors-versioning

Revision ID: 976f15a37e6a
Revises: 6e834843e7e9
Create Date: 2022-08-19 11:22:52.085339

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import text

import mindsdb.interfaces.storage.db    # noqa


# revision identifiers, used by Alembic.
revision = '976f15a37e6a'
down_revision = '6e834843e7e9'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('deleted_at', sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column('active', sa.Boolean(), nullable=True))

    conn.execute(text('''
        update predictor set active = :val;
    '''), {
        'val': True
    })

    session.commit()


def downgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_column('active')
        batch_op.drop_column('deleted_at')



import os

import pandas as pd
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build

from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.libs.api_handler import APIHandler, FuncParser
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log
logger = log.getLogger(__name__)

from mindsdb.integrations.handlers.gmail_handler.utils import AuthException, google_auth_flow, save_creds_to_file

from .google_calendar_tables import GoogleCalendarEventsTable


class GoogleCalendarHandler(APIHandler):
    """
        A class for handling connections and interactions with the Google Calendar API.
    """
    name = 'google_calendar'

    def __init__(self, name: str, **kwargs):
        """ constructor
        Args:
            name (str): the handler name
            credentials_file (str): The path to the credentials file.
            scopes (list): The list of scopes to use for authentication.
            is_connected (bool): Whether the API client is connected to Google Calendar.
            events (GoogleCalendarEventsTable): The `GoogleCalendarEventsTable` object for interacting with the events table.
        """
        super().__init__(name)

        self.token = None
        self.service = None
        self.connection_data = kwargs.get('connection_data', {})
        self.credentials_file = self.connection_data['credentials']
        self.scopes = [
            'https://www.googleapis.com/auth/calendar',
            'https://www.googleapis.com/auth/calendar.events',
            'https://www.googleapis.com/auth/calendar.readonly'
        ]
        self.is_connected = False

        self.handler_storage = kwargs['handler_storage']

        events = GoogleCalendarEventsTable(self)
        self.events = events
        self._register_table('events', events)

    def connect(self):
        """
        Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            HandlerStatusResponse
        """
        if self.is_connected is True:
            return self.service

        secret_file = self.credentials_file

        curr_dir = self.handler_storage.folder_get('config')

        creds_file = None
        try:
            creds_file = os.path.join(curr_dir, 'secret.json')
        except Exception:
            pass

        creds = None
        if os.path.isfile(creds_file):
            creds = Credentials.from_authorized_user_file(creds_file, self.scopes)

        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())

                save_creds_to_file(creds, creds_file)
                self.handler_storage.folder_sync('config')

            else:
                creds = google_auth_flow(secret_file, self.scopes, self.connection_data.get('code'))

                save_creds_to_file(creds, creds_file)
                self.handler_storage.folder_sync('config')

        self.service = build('calendar', 'v3', credentials=creds)
        return self.service

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True

        except AuthException as error:
            response.error_message = str(error)
            response.redirect_url = error.auth_url
            return response

        except Exception as e:
            logger.error(f'Error connecting to Google Calendar API: {e}!')
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str = None) -> Response:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, api's json etc)
        Returns:
            HandlerResponse
        """
        method_name, params = FuncParser().from_string(query)

        df = self.call_application_api(method_name, params)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def get_events(self, params: dict = None) -> pd.DataFrame:
        """
        Get events from Google Calendar API
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        page_token = None
        events = pd.DataFrame(columns=self.events.get_columns())
        while True:
            events_result = service.events().list(calendarId='primary', pageToken=page_token, **params).execute()
            events = pd.concat(
                [events, pd.DataFrame(events_result.get('items', []), columns=self.events.get_columns())],
                ignore_index=True
            )
            page_token = events_result.get('nextPageToken')
            if not page_token:
                break
        return events

    def create_event(self, params: dict = None) -> pd.DataFrame:
        """
        Create an event in the calendar.
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        # Check if 'attendees' is a string and split it into a list
        if isinstance(params['attendees'], str):
            params['attendees'] = params['attendees'].split(',')

        event = {
            'summary': params['summary'],
            'location': params['location'],
            'description': params['description'],
            'start': {
                'dateTime': params['start']['dateTime'],
                'timeZone': params['start']['timeZone'],
            },
            'end': {
                'dateTime': params['end']['dateTime'],
                'timeZone': params['end']['timeZone'],
            },
            'recurrence': [
                'RRULE:FREQ=DAILY;COUNT=1'
            ],
            'attendees': [{'email': attendee['email']} for attendee in (params['attendees'] 
                            if isinstance(params['attendees'], list) else [params['attendees']])],
            'reminders': {
                'useDefault': False,
                'overrides': [
                    {'method': 'email', 'minutes': 24 * 60},
                    {'method': 'popup', 'minutes': 10},
                ],
            },
        }

        event = service.events().insert(calendarId='primary', 
                                        body=event).execute()
        return pd.DataFrame([event], columns=self.events.get_columns())

    def update_event(self, params: dict = None) -> pd.DataFrame:
        """
        Update event or events in the calendar.
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        df = pd.DataFrame(columns=['eventId', 'status'])
        if params['event_id']:
            start_id = int(params['event_id'])
            end_id = start_id + 1
        elif not params['start_id']:
            start_id = int(params['end_id']) - 10
        elif not params['end_id']:
            end_id = int(params['start_id']) + 10
        else:
            start_id = int(params['start_id'])
            end_id = int(params['end_id'])

        for i in range(start_id, end_id):
            event = service.events().get(calendarId='primary', eventId=i).execute()
            if params['summary']:
                event['summary'] = params['summary']
            if params['location']:
                event['location'] = params['location']
            if params['description']:
                event['description'] = params['description']
            if params['start']:
                event['start']['dateTime'] = params['start']['dateTime']
                event['start']['timeZone'] = params['start']['timeZone']
            if params['end']:
                event['end']['dateTime'] = params['end']['dateTime']
                event['end']['timeZone'] = params['end']['timeZone']
            if params['attendees']:
                event['attendees'] = [{'email': attendee} for attendee in params['attendees'].split(',')]
            updated_event = service.events().update(calendarId='primary', eventId=event['id'], body=event).execute()
            df = pd.concat([df, pd.DataFrame([{'eventId': updated_event['id'], 'status': 'updated'}])],
                           ignore_index=True)

        return df

    def delete_event(self, params):
        """
        Delete event or events in the calendar.
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        if params['event_id']:
            service.events().delete(calendarId='primary', eventId=params['event_id']).execute()
            return pd.DataFrame([{'eventId': params['event_id'], 'status': 'deleted'}])
        else:
            df = pd.DataFrame(columns=['eventId', 'status'])
            if not params['start_id']:
                start_id = int(params['end_id']) - 10
            elif not params['end_id']:
                end_id = int(params['start_id']) + 10
            else:
                start_id = int(params['start_id'])
                end_id = int(params['end_id'])
            for i in range(start_id, end_id):
                service.events().delete(calendarId='primary', eventId=str(i)).execute()
                df = pd.concat([df, pd.DataFrame([{'eventId': str(i), 'status': 'deleted'}])], ignore_index=True)
            return df

    def call_application_api(self, method_name: str = None, params: dict = None) -> pd.DataFrame:
        """
        Call Google Calendar API and map the data to pandas DataFrame
        Args:
            method_name (str): method name
            params (dict): query parameters
        Returns:
            DataFrame
        """
        if method_name == 'get_events':
            return self.get_events(params)
        elif method_name == 'create_event':
            return self.create_event(params)
        elif method_name == 'update_event':
            return self.update_event(params)
        elif method_name == 'delete_event':
            return self.delete_event(params)
        else:
            raise NotImplementedError(f'Unknown method {method_name}')



import time
import hmac
import base64
import hashlib
import datetime
from typing import Dict

import pandas as pd
import requests

from mindsdb.integrations.handlers.coinbase_handler.coinbase_tables import CoinBaseAggregatedTradesTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)

from mindsdb_sql import parse_sql

_BASE_COINBASE_US_URL = 'https://api.exchange.coinbase.com'


class CoinBaseHandler(APIHandler):
    """A class for handling connections and interactions with the CoinBase API.

    Attributes:
        api_key (str): API key
        api_secret (str): API secret
        is_connected (bool): Whether or not the API client is connected to CoinBase.

    """

    def __init__(self, name: str = None, **kwargs):
        """Registers all API tables and prepares the handler for an API connection.

        Args:
            name: (str): The handler name to use
        """
        super().__init__(name)
        self.api_key = None
        self.api_secret = None
        self.api_passphrase = None

        args = kwargs.get('connection_data', {})
        if 'api_key' in args:
            self.api_key = args['api_key']
        if 'api_secret' in args:
            self.api_secret = args['api_secret']
        if 'api_passphrase' in args:
            self.api_passphrase = args['api_passphrase']
        self.client = None
        self.is_connected = False

        coinbase_candle_data = CoinBaseAggregatedTradesTable(self)
        self._register_table('coinbase_candle_data', coinbase_candle_data)

    def connect(self):
        """Creates a new CoinBase API client if needed and sets it as the client to use for requests.

        Returns newly created CoinBase API client, or current client if already set.
        """
        self.is_connected = True
        return self.client

    def check_connection(self) -> StatusResponse:
        """Checks connection to CoinBase API by sending a ping request.

        Returns StatusResponse indicating whether or not the handler is connected.
        """
        response = StatusResponse(True)
        self.is_connected = response.success
        return response

    # symbol: BTC-USD
    # granularity 60, 300, 900, 3600, 21600, 86400
    def get_coinbase_candle(self, symbol: str, granularity: int) -> pd.DataFrame:
        jdocs = []
        current_time = datetime.datetime.now()
        start_time = current_time - datetime.timedelta(seconds=granularity)
        start_time_iso = start_time.isoformat().split(".")[0] + "-04:00"
        path = "/products/" + symbol + "/candles?granularity=" + str(granularity) + "&start=" + start_time_iso
        headers = self.generate_api_headers("GET", path)
        url = _BASE_COINBASE_US_URL + path
        response = requests.get(url, headers=headers)
        candles = response.json()
        for candle in candles:
            dt = datetime.datetime.fromtimestamp(candle[0], None).isoformat()
            low, high, open, close, volume = candle[1:]
            jdoc = {"symbol": symbol, "low": low, "high": high, "open": open, "close": close, "volume": volume, "timestamp": candle[0], "timestamp_iso": dt}
            jdocs.append(jdoc)
        return pd.DataFrame(jdocs)

    def _get_candle(self, params: Dict = None) -> pd.DataFrame:
        """Gets aggregate trade data for a symbol based on given parameters

        Returns results as a pandas DataFrame.

        Args:
            params (Dict): Trade data params (symbol, interval)
        """
        if 'symbol' not in params:
            raise ValueError('Missing "symbol" param to fetch trade data for.')
        if 'interval' not in params:
            raise ValueError('Missing "interval" param (60, 300, 900, 3600, 21600, 86400).')

        candle = self.get_coinbase_candle(params['symbol'], int(params['interval']))
        return candle

    def native_query(self, query: str = None) -> Response:
        ast = parse_sql(query, dialect='mindsdb')
        return self.query(ast)

    def generate_api_headers(self, method: str, path: str) -> dict:
        timestamp = str(int(time.time()))
        message = timestamp + method + path
        signature = base64.b64encode(hmac.new(base64.b64decode(self.api_secret), str.encode(message), hashlib.sha256).digest())
        headers = {
            "Content-Type": "application/json",
            "CB-ACCESS-SIGN": signature,
            "CB-ACCESS-KEY": self.api_key,
            "CB-ACCESS-TIMESTAMP": timestamp,
            "CB-VERSION": "2015-04-08",
            "CB-ACCESS-PASSPHRASE": self.api_passphrase
        }
        return headers

    def call_coinbase_api(self, method_name: str = None, params: Dict = None) -> pd.DataFrame:
        """Calls the CoinBase API method with the given params.

        Returns results as a pandas DataFrame.

        Args:
            method_name (str): Method name to call
            params (Dict): Params to pass to the API call
        """
        if method_name == 'get_candle':
            return self._get_candle(params)
        raise NotImplementedError('Method name {} not supported by CoinBase API Handler'.format(method_name))



import os
import shutil
import tarfile
import tempfile
import zipfile

import multipart
import requests
from flask import current_app as ca
from flask import request
from flask_restx import Resource

from mindsdb.api.http.namespaces.configs.files import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities import log
from mindsdb.utilities.security import is_private_url, clear_filename
from mindsdb.utilities.fs import safe_extract

logger = log.getLogger(__name__)


@ns_conf.route("/")
class FilesList(Resource):
    @ns_conf.doc("get_files_list")
    @api_endpoint_metrics('GET', '/files')
    def get(self):
        """List all files"""
        return ca.file_controller.get_files()


@ns_conf.route("/<name>")
@ns_conf.param("name", "MindsDB's name for file")
class File(Resource):
    @ns_conf.doc("put_file")
    @api_endpoint_metrics('PUT', '/files/file')
    def put(self, name: str):
        """add new file
        params in FormData:
            - file
            - original_file_name [optional]
        """

        data = {}
        mindsdb_file_name = name

        existing_file_names = ca.file_controller.get_files_names()

        def on_field(field):
            name = field.field_name.decode()
            value = field.value.decode()
            data[name] = value

        file_object = None

        def on_file(file):
            nonlocal file_object
            data["file"] = file.file_name.decode()
            file_object = file.file_object

        temp_dir_path = tempfile.mkdtemp(prefix="mindsdb_file_")

        if request.headers["Content-Type"].startswith("multipart/form-data"):
            parser = multipart.create_form_parser(
                headers=request.headers,
                on_field=on_field,
                on_file=on_file,
                config={
                    "UPLOAD_DIR": temp_dir_path.encode(),  # bytes required
                    "UPLOAD_KEEP_FILENAME": True,
                    "UPLOAD_KEEP_EXTENSIONS": True,
                    "MAX_MEMORY_FILE_SIZE": 0,
                },
            )

            while True:
                chunk = request.stream.read(8192)
                if not chunk:
                    break
                parser.write(chunk)
            parser.finalize()
            parser.close()

            if file_object is not None and not file_object.closed:
                file_object.close()
        else:
            data = request.json

        if mindsdb_file_name in existing_file_names:
            return http_error(
                400,
                "File already exists",
                f"File with name '{data['file']}' already exists",
            )

        if data.get("source_type") == "url":
            url = data["source"]
            data["file"] = clear_filename(data["name"])

            config = Config()
            is_cloud = config.get("cloud", False)
            if is_cloud and is_private_url(url):
                return http_error(
                    400, f'URL is private: {url}'
                )

            if is_cloud is True and ctx.user_class != 1:
                info = requests.head(url)
                file_size = info.headers.get("Content-Length")
                try:
                    file_size = int(file_size)
                except Exception:
                    pass

                if file_size is None:
                    return http_error(
                        400,
                        "Error getting file info",
                        "Сan't determine remote file size",
                    )
                if file_size > 1024 * 1024 * 100:
                    return http_error(
                        400, "File is too big", "Upload limit for file is 100Mb"
                    )
            with requests.get(url, stream=True) as r:
                if r.status_code != 200:
                    return http_error(
                        400, "Error getting file", f"Got status code: {r.status_code}"
                    )
                file_path = os.path.join(temp_dir_path, data["file"])
                with open(file_path, "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)

        original_file_name = clear_filename(data.get("original_file_name"))

        file_path = os.path.join(temp_dir_path, data["file"])
        lp = file_path.lower()
        if lp.endswith((".zip", ".tar.gz")):
            if lp.endswith(".zip"):
                with zipfile.ZipFile(file_path) as f:
                    f.extractall(temp_dir_path)
            elif lp.endswith(".tar.gz"):
                with tarfile.open(file_path) as f:
                    safe_extract(f, temp_dir_path)
            os.remove(file_path)
            files = os.listdir(temp_dir_path)
            if len(files) != 1:
                os.rmdir(temp_dir_path)
                return http_error(
                    400, "Wrong content.", "Archive must contain only one data file."
                )
            file_path = os.path.join(temp_dir_path, files[0])
            mindsdb_file_name = files[0]
            if not os.path.isfile(file_path):
                os.rmdir(temp_dir_path)
                return http_error(
                    400, "Wrong content.", "Archive must contain data file in root."
                )

        try:
            ca.file_controller.save_file(
                mindsdb_file_name, file_path, file_name=original_file_name
            )
        except Exception as e:
            return http_error(500, 'Error', str(e))
        finally:
            shutil.rmtree(temp_dir_path, ignore_errors=True)

        return "", 200

    @ns_conf.doc("delete_file")
    @api_endpoint_metrics('DELETE', '/files/file')
    def delete(self, name: str):
        """delete file"""

        try:
            ca.file_controller.delete_file(name)
        except Exception as e:
            logger.error(e)
            return http_error(
                400,
                "Error deleting file",
                f"There was an error while tring to delete file with name '{name}'",
            )
        return "", 200



import unittest

from mindsdb.integrations.handlers.ckan_handler.ckan_handler import CkanHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class CkanHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "url": "http://demo.ckan.org/"}
        }
        cls.handler = CkanHandler('test_ckan_handler', **cls.kwargs)

    def test_connect(self):
        self.handler.connect()

    def test_query(self):
        self.handler.query("SELECT * from 'b53c9e72-6b59-4cda-8c0c-7d6a51dad12a'")

    def test_disconnect(self):
        self.handler.disconnect()

    def test_get_tables(self):
        tables = self.handler.get_tables()
        assert tables['type'] is not RESPONSE_TYPE.ERROR



from collections import defaultdict

from mindsdb_sql.parser.ast import (
    Identifier,
    Select,
    Star
)
from mindsdb_sql.planner.steps import SubSelectStep, QueryStep
from mindsdb_sql.planner.utils import query_traversal

from mindsdb.api.executor.sql_query.result_set import ResultSet, Column
from mindsdb.api.executor.utilities.sql import query_df

from mindsdb.interfaces.query_context.context_controller import query_context_controller

from mindsdb.api.executor.exceptions import KeyColumnDoesNotExist

from .base import BaseStepCall
from .fetch_dataframe import get_fill_param_fnc


class SubSelectStepCall(BaseStepCall):

    bind = SubSelectStep

    def call(self, step):
        result = self.steps_data[step.dataframe.step_num]

        table_name = step.table_name
        if table_name is None:
            table_name = 'df_table'
        else:
            table_name = table_name

        query = step.query
        query.from_table = Identifier('df_table')

        if step.add_absent_cols and isinstance(query, Select):
            query_cols = set()

            def f_all_cols(node, **kwargs):
                if isinstance(node, Identifier):
                    query_cols.add(node.parts[-1])

            query_traversal(query.where, f_all_cols)

            result_cols = [col.name for col in result.columns]

            for col_name in query_cols:
                if col_name not in result_cols:
                    result.add_column(Column(col_name))

        df = result.to_df()
        res = query_df(df, query)

        result2 = ResultSet()
        # get database from first column
        database = result.columns[0].database
        result2.from_df(res, database, table_name)

        return result2


class QueryStepCall(BaseStepCall):

    bind = QueryStep

    def call(self, step):
        query = step.query

        if step.from_table is not None:
            result_set = self.steps_data[step.from_table.step_num]
        else:
            # only from_table can content result
            prev_step_num = query.from_table.value.step_num
            result_set = self.steps_data[prev_step_num]

        df, col_names = result_set.to_df_cols()
        col_idx = {}
        tbl_idx = defaultdict(list)
        for name, col in col_names.items():
            col_idx[col.alias] = name
            col_idx[(col.table_alias, col.alias)] = name
            # add to tables
            tbl_idx[col.table_name].append(name)
            if col.table_name != col.table_alias:
                tbl_idx[col.table_alias].append(name)

        # analyze condition and change name of columns
        def check_fields(node, is_target=None, **kwargs):

            if isinstance(node, Identifier):
                # only column name
                col_name = node.parts[-1]
                if is_target and isinstance(col_name, Star):
                    if len(node.parts) == 1:
                        # left as is
                        return
                    else:
                        # replace with all columns from table
                        table_name = node.parts[-2]
                        return [
                            Identifier(parts=[col])
                            for col in tbl_idx.get(table_name, [])
                        ]

                if len(node.parts) == 1:
                    key = col_name
                else:
                    table_name = node.parts[-2]
                    key = (table_name, col_name)

                if key not in col_idx:
                    raise KeyColumnDoesNotExist(f'Table not found for column: {key}')

                new_name = col_idx[key]
                return Identifier(parts=[new_name], alias=node.alias)

        # fill params
        fill_params = get_fill_param_fnc(self.steps_data)
        query_traversal(query, fill_params)

        query_traversal(query, check_fields)
        query.where = query_context_controller.remove_lasts(query.where)

        query.from_table = Identifier('df_table')
        res = query_df(df, query)

        data = ResultSet().from_df_cols(res, col_names, strict=False)

        return data



from pydantic import BaseModel, AnyUrl, model_validator
from urllib.parse import urlparse


class ConnectionConfig(BaseModel):
    # TODO: For now validate AnyURL since MySQLDsn wasn't working
    url: AnyUrl = None
    host: str = None
    port: int = 3306
    user: str = None
    password: str = None
    database: str = None

    @model_validator(mode="before")
    def check_db_params(cls, values):
        """Ensures either URL is provided or all individual parameters are provided."""
        url = values.get('url')
        host = values.get('host')
        user = values.get('user')
        password = values.get('password')
        database = values.get('database')
        if not url and not (host and user and password and database):
            raise ValueError("Either a valid URL or required parameters (host, user, password, database) must be provided.")

        if url:
            parsed = urlparse(url)
            values['host'] = parsed.hostname or host
            values['port'] = parsed.port if parsed.port is not None else 3306
            values['user'] = parsed.username or user
            values['password'] = parsed.password or password
            values['database'] = parsed.path[1:] if parsed.path else database

            # mysql connector raise error if url is provided
            values.pop('url', None)

            return values

        if not url:
            for param in ['host', 'user', 'password', 'database']:
                if not values.get(param):
                    raise ValueError(f"{param} is required when URL is not provided.")
        return values

    class Config:
        str_min_length = 1
        str_strip_whitespace = True



from http import HTTPStatus
import functools
import time

from prometheus_client import Histogram, Summary


INTEGRATION_HANDLER_QUERY_TIME = Summary(
    'mindsdb_integration_handler_query_seconds',
    'How long integration handlers take to answer queries',
    ('integration', 'response_type')
)

INTEGRATION_HANDLER_RESPONSE_SIZE = Summary(
    'mindsdb_integration_handler_response_size',
    'How many rows are returned by an integration handler query',
    ('integration', 'response_type')
)

_REST_API_LATENCY = Histogram(
    'mindsdb_rest_api_latency_seconds',
    'How long REST API requests take to complete, grouped by method, endpoint, and status',
    ('method', 'endpoint', 'status')
)


def api_endpoint_metrics(method: str, uri: str):
    def decorator_metrics(endpoint_func):
        @functools.wraps(endpoint_func)
        def wrapper_metrics(*args, **kwargs):
            time_before_query = time.perf_counter()
            try:
                response = endpoint_func(*args, **kwargs)
            except Exception as e:
                # Still record metrics for unexpected exceptions.
                elapsed_seconds = time.perf_counter() - time_before_query
                api_latency_with_labels = _REST_API_LATENCY.labels(
                    method, uri, HTTPStatus.INTERNAL_SERVER_ERROR.value)
                api_latency_with_labels.observe(elapsed_seconds)
                raise e
            elapsed_seconds = time.perf_counter() - time_before_query
            status = response.status_code if hasattr(response, 'status_code') else HTTPStatus.OK.value
            api_latency_with_labels = _REST_API_LATENCY.labels(method, uri, status)
            api_latency_with_labels.observe(elapsed_seconds)
            return response
        return wrapper_metrics
    return decorator_metrics



import json
import os
import shutil
import tempfile
from io import BytesIO, StringIO
from unittest.mock import patch

import pandas
import pytest
import responses
from mindsdb_sql.exceptions import ParsingException
from mindsdb_sql.parser.ast import CreateTable, DropTables, Identifier, Select, Star
from pytest_lazyfixture import lazy_fixture

from mindsdb.integrations.handlers.file_handler.file_handler import FileHandler
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.interfaces.file.file_controller import FileController

# Define a table to use as content for all of the file types
# This data needs to match that saved in the files in the ./data/ dir (except pdf and txt files)
test_file_content = [
    ["col_one", "col_two", "col_three", "col_four"],
    [1, -1, 0.1, "A"],
    [2, -2, 0.2, "B"],
    [3, -3, 0.3, "C"],
]

file_records = [("one", 1, test_file_content[0]), ("two", 2, test_file_content[0])]


class MockFileController:
    """
    Pretends to be a file controller. Gives details of 'file_records' above, and mocks file deletion.
    We're not testing the file controller here, so we don't need to rely on it.
    """

    def get_files(self):
        return [
            {
                "name": record[0],
                "row_count": record[1],
                "columns": record[2],
            }
            for record in file_records
        ]

    def get_file_meta(self, *args, **kwargs):
        return self.get_files()[0]

    def delete_file(self, name):
        return True


def curr_dir():
    return os.path.dirname(os.path.realpath(__file__))


# Fixtures to get a path to a partiular type of file
@pytest.fixture
def csv_file() -> str:
    return os.path.join(curr_dir(), "data", "test.csv")


@pytest.fixture
def xlsx_file() -> str:
    return os.path.join(curr_dir(), "data", "test.xlsx")


@pytest.fixture
def json_file() -> str:
    return os.path.join(curr_dir(), "data", "test.json")


@pytest.fixture
def parquet_file() -> str:
    return os.path.join(curr_dir(), "data", "test.parquet")


@pytest.fixture
def pdf_file() -> str:
    return os.path.join(curr_dir(), "data", "test.pdf")


@pytest.fixture
def txt_file() -> str:
    return os.path.join(curr_dir(), "data", "test.txt")


class TestIsItX:
    """Tests all of the 'is_it_x()' functions to determine a file's type"""

    # We can't test xlsx or parquet here because they're binary files
    @pytest.mark.parametrize(
        "file_path,result",
        [(lazy_fixture("csv_file"), True), (lazy_fixture("json_file"), False)],
    )
    def test_is_it_csv(self, file_path, result):
        with open(file_path, "r") as fh:
            assert FileHandler.is_it_csv(StringIO(fh.read())) is result

    @pytest.mark.parametrize(
        "file_path,result",
        [
            (lazy_fixture("csv_file"), False),
            (lazy_fixture("xlsx_file"), True),
            (lazy_fixture("json_file"), False),
            (lazy_fixture("parquet_file"), False),
            (lazy_fixture("txt_file"), False),
            (lazy_fixture("pdf_file"), False),
        ],
    )
    def test_is_it_xlsx(self, file_path, result):
        assert FileHandler.is_it_xlsx(file_path) is result

    # We can't test xlsx or parquet here because they're binary files
    @pytest.mark.parametrize(
        "file_path,result",
        [
            (lazy_fixture("csv_file"), False),
            (lazy_fixture("json_file"), True),
            (lazy_fixture("txt_file"), False),
        ],
    )
    def test_is_it_json(self, file_path, result):
        with open(file_path, "r") as fh:
            assert FileHandler.is_it_json(StringIO(fh.read())) is result

    @pytest.mark.parametrize(
        "file_path,result",
        [
            (lazy_fixture("csv_file"), False),
            (lazy_fixture("xlsx_file"), False),
            (lazy_fixture("json_file"), False),
            (lazy_fixture("parquet_file"), True),
            (lazy_fixture("txt_file"), False),
            (lazy_fixture("pdf_file"), False),
        ],
    )
    def test_is_it_parquet(self, file_path, result):
        with open(file_path, "rb") as fh:
            assert FileHandler.is_it_parquet(BytesIO(fh.read())) is result


class TestQuery:
    """Tests all of the scenarios relating to the query() function"""

    def test_query_drop(self, monkeypatch):
        """Test a valid drop table query"""

        def mock_delete(self, name):
            return True

        monkeypatch.setattr(MockFileController, "delete_file", mock_delete)
        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.query(DropTables([Identifier(parts=["one"])]))

        assert response.type == RESPONSE_TYPE.OK

    def test_query_drop_bad_delete(self, monkeypatch):
        """Test an invalid drop table query"""

        def mock_delete(self, name):
            raise Exception("File delete error")

        monkeypatch.setattr(MockFileController, "delete_file", mock_delete)
        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.query(DropTables([Identifier(parts=["one"])]))

        assert response.type == RESPONSE_TYPE.ERROR

    def test_query_select(self, csv_file):
        """Test a valid select query"""
        expected_df = pandas.read_csv(csv_file)

        # This is temporary because the file controller currently absconds with our file when we save it:
        # https://github.com/mindsdb/mindsdb/issues/8141
        csv_tmp = os.path.join(tempfile.gettempdir(), "test.csv")
        if os.path.exists(csv_tmp):
            os.remove(csv_tmp)
        shutil.copy(csv_file, csv_tmp)

        # Configure mindsdb and set up the file controller
        # Ideally this would be a lot simpler..
        db_file = tempfile.mkstemp(prefix="mindsdb_db_")[1]
        config = {"storage_db": "sqlite:///" + db_file}
        fdi, cfg_file = tempfile.mkstemp(prefix="mindsdb_conf_")
        with os.fdopen(fdi, "w") as fd:
            json.dump(config, fd)
        os.environ["MINDSDB_CONFIG_PATH"] = cfg_file

        from mindsdb.utilities.config import Config

        Config()
        from mindsdb.interfaces.storage import db

        db.init()
        db.session.rollback()
        db.Base.metadata.drop_all(db.engine)

        # create
        db.Base.metadata.create_all(db.engine)

        # fill with data
        r = db.Integration(name="files", data={}, engine="files")
        db.session.add(r)
        db.session.flush()
        # Config #

        file_controller = FileController()
        file_controller.save_file(
            os.path.splitext(os.path.basename(csv_file))[0], csv_tmp
        )

        file_handler = FileHandler(file_controller=file_controller)
        response = file_handler.query(
            Select(
                targets=[Star()],
                from_table=Identifier(
                    parts=[os.path.splitext(os.path.basename(csv_file))[0]]
                ),
            )
        )

        assert response.type == RESPONSE_TYPE.TABLE
        assert response.error_code == 0
        assert response.error_message is None
        assert expected_df.equals(response.data_frame)

    def test_query_bad_type(self):
        """Test an invalid query type for files"""
        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.query(CreateTable([Identifier(parts=["someTable"])]))

        assert response.type == RESPONSE_TYPE.ERROR

    def test_native_query(self, monkeypatch):
        """Test a valid native table query"""

        def mock_delete(self, name):
            return True

        monkeypatch.setattr(MockFileController, "delete_file", mock_delete)
        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.native_query("DROP TABLE one")

        assert response.type == RESPONSE_TYPE.OK

    def test_invalid_native_query(self):
        file_handler = FileHandler(file_controller=MockFileController())
        with pytest.raises(ParsingException):
            file_handler.native_query("INVALID QUERY")


def test_get_file_path_with_file_path():
    """Test an valid native table query"""
    file_path = "example.txt"
    result = FileHandler._get_file_path(file_path)
    assert result == file_path


@patch("mindsdb.integrations.handlers.file_handler.file_handler.FileHandler._fetch_url")
def test_get_file_path_with_url(mock_fetch_url):
    url = "http://example.com/file.txt"
    expected_result = "some_file_path"
    # we test _fetch_url separately below. Mock it for this test
    mock_fetch_url.return_value = expected_result

    result = FileHandler._get_file_path(url)

    assert result == expected_result
    mock_fetch_url.assert_called_with(url)


@pytest.mark.parametrize(
    "file_path,expected_columns",
    [
        (lazy_fixture("csv_file"), test_file_content[0]),
        (lazy_fixture("xlsx_file"), test_file_content[0]),
        (lazy_fixture("json_file"), test_file_content[0]),
        (lazy_fixture("parquet_file"), test_file_content[0]),
        (lazy_fixture("pdf_file"), ["content", "metadata"]),
        (lazy_fixture("txt_file"), ["content", "metadata"]),
    ],
)
def test_handle_source(file_path, expected_columns):
    df, col_map = FileHandler._handle_source(file_path)
    assert isinstance(df, pandas.DataFrame)
    assert df.columns.tolist() == expected_columns

    # The pdf and txt files have some different content
    if not file_path.endswith(".pdf") and not file_path.endswith(".txt"):
        assert len(df) == len(test_file_content) - 1
        assert df.values.tolist() == test_file_content[1:]


@pytest.mark.parametrize(
    "file_path,expected_file_type,expected_delimiter,expected_data_type",
    [
        (lazy_fixture("csv_file"), "csv", ",", StringIO),
        (lazy_fixture("xlsx_file"), "xlsx", None, BytesIO),
        (lazy_fixture("json_file"), "json", None, StringIO),
        (lazy_fixture("parquet_file"), "parquet", None, BytesIO),
        (lazy_fixture("pdf_file"), "pdf", None, BytesIO),
        (lazy_fixture("txt_file"), "txt", None, BytesIO),
    ],
)
def test_get_data_io(
    file_path, expected_file_type, expected_delimiter, expected_data_type
):
    data_io, file_type, file_dialect = FileHandler._get_data_io(file_path)
    assert file_type == expected_file_type
    assert type(data_io) == expected_data_type
    if expected_delimiter is None:
        assert file_dialect is None
    else:
        assert file_dialect.delimiter == expected_delimiter


@pytest.mark.parametrize(
    "csv_string,delimiter",
    [
        (StringIO("example,csv,file"), ","),
        (StringIO("example;csv;file"), ";"),
        (StringIO("example\tcsv\tfile"), "\t"),
    ],
)
def test_check_valid_dialects(csv_string, delimiter):
    dialect = FileHandler._get_csv_dialect(csv_string)
    assert dialect.delimiter == delimiter


def test_check_invalid_dialects():
    with pytest.raises(Exception):
        FileHandler._get_csv_dialect("example csv file")
    with pytest.raises(Exception):
        FileHandler._get_csv_dialect("example\ncsv\nfile")
    with pytest.raises(Exception):
        FileHandler._get_csv_dialect("example|csv|file")


@responses.activate
def test_fetch_url():
    file_content = "Fake File Content 1234567890"
    file_url = "https://test.fake/robots.txt"
    responses.add(
        responses.GET, file_url, body=file_content, status=200
    )  # mock the response

    file_path = FileHandler._fetch_url(file_url)
    with open(file_path, "r") as fh:
        saved_file_content = fh.read()

    assert saved_file_content == file_content


@responses.activate
def test_fetch_url_raises():
    responses.add(responses.GET, "https://google.com", status=404)

    with pytest.raises(Exception):
        FileHandler._fetch_url("obvious_broken_url")
    with pytest.raises(Exception):
        FileHandler._fetch_url("https://google.com")  # will get 404 response


def test_get_tables():
    file_handler = FileHandler(file_controller=MockFileController())
    response = file_handler.get_tables()

    assert response.type == RESPONSE_TYPE.TABLE

    expected_df = pandas.DataFrame(
        [
            {"TABLE_NAME": x[0], "TABLE_ROWS": x[1], "TABLE_TYPE": "BASE TABLE"}
            for x in file_records
        ]
    )

    assert response.data_frame.equals(expected_df)


def test_get_columns():
    file_handler = FileHandler(file_controller=MockFileController())
    response = file_handler.get_columns("mock")

    assert response.type == RESPONSE_TYPE.TABLE

    expected_df = pandas.DataFrame(
        [{"Field": x, "Type": "str"} for x in file_records[0][2]]
    )

    assert response.data_frame.equals(expected_df)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The IP address/host name of the Teradata instance host.'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the user name.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'Specifies the password for the user.',
        'secret': True
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the initial database to use after logon, instead of the default database.'
    },
    dbs_port={
        'type': ARG_TYPE.STR,
        'description': 'The port number of the Teradata instance.'
    },
    encryptdata={
        'type': ARG_TYPE.STR,
        'description': 'Controls encryption of data exchanged between the driver and the database.'
    },
    https_port={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the database port number for HTTPS/TLS connections.'
    },
    sslca={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the file name of a PEM file that contains Certificate Authority (CA) certificates.'
    },
    sslcapath={
        'type': ARG_TYPE.STR,
        'description': 'Specifies a directory of PEM files that contain Certificate Authority (CA) certificates.'
    },
    sslcipher={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the TLS cipher for HTTPS/TLS connections.'
    },
    sslmode={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the SSL mode for HTTPS/TLS connections.'
    },
    sslprotocol={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the TLS protocol for HTTPS/TLS connections.'
    },
    tmode={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the Teradata transaction mode.'
    },
    logmech={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the login authentication method.'
    },
    logdata={
        'type': ARG_TYPE.STR,
        'description': 'Specifies extra data for the chosen logon authentication method.'
    },
    browser={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the command to open the browser for Browser Authentication.'
    }
)

connection_args_example = OrderedDict(
    host='192.168.0.41',
    user='dbc',
    password='dbc',
    database='HR'
)



from typing import List, Optional

import pandas as pd
from pymilvus import Collection, CollectionSchema, DataType, FieldSchema, connections, utility

from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.integrations.libs.vectordatabase_handler import FilterCondition, FilterOperator, TableField, VectorStoreHandler
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MilvusHandler(VectorStoreHandler):
    """This handler handles connection and execution of the Milvus statements."""

    name = "milvus"

    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self._connection_data = kwargs["connection_data"]
        # Extract parameters used while searching and leave the rest for establishing connection
        self._search_limit = 100
        if "search_default_limit" in self._connection_data:
            self._search_limit = self._connection_data["search_default_limit"]
        self._search_params = {
            "search_metric_type": "L2",
            "search_ignore_growing": False,
            "search_params": {"nprobe": 10},
        }
        for search_param_name in self._search_params:
            if search_param_name in self._connection_data:
                self._search_params[search_param_name] = self._connection_data[search_param_name]
        # Extract parameters used for creating tables
        self._create_table_params = {
            "create_auto_id": False,
            "create_id_max_len": 64,
            "create_embedding_dim": 8,
            "create_dynamic_field": True,
            "create_content_max_len": 200,
            "create_content_default_value": "",
            "create_schema_description": "MindsDB generated table",
            "create_alias": "default",
            "create_index_params": {},
            "create_index_metric_type": "L2",
            "create_index_type": "AUTOINDEX",
        }
        for create_table_param in self._create_table_params:
            if create_table_param in self._connection_data:
                self._create_table_params[create_table_param] = self._connection_data[create_table_param]
        self.is_connected = False
        self.connect()

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """Connect to a Milvus database."""
        if self.is_connected is True:
            return
        try:
            connections.connect(**self._connection_data)
            self.is_connected = True
        except Exception as e:
            logger.error(f"Error connecting to Milvus client: {e}!")
            self.is_connected = False

    def disconnect(self):
        """Close the database connection."""
        if self.is_connected is False:
            return
        connections.disconnect(self._connection_data["alias"])
        self.is_connected = False

    def check_connection(self):
        """Check the connection to the Milvus database."""
        response_code = StatusResponse(False)
        try:
            response_code.success = connections.has_connection(self._connection_data["alias"])
        except Exception as e:
            logger.error(f"Error checking Milvus connection: {e}!")
            response_code.error_message = str(e)
        return response_code

    def get_tables(self) -> HandlerResponse:
        """Get the list of collections in the Milvus database."""
        collections = utility.list_collections()
        collections_name = pd.DataFrame(
            columns=["TABLE_NAME"],
            data=[collection for collection in collections],
        )
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=collections_name)

    def drop_table(self, table_name: str, if_exists=True) -> HandlerResponse:
        """Delete a collection from the Milvus database."""
        try:
            utility.drop_collection(table_name)
        except Exception as e:
            if if_exists:
                return Response(resp_type=RESPONSE_TYPE.OK)
            else:
                return Response(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_message=f"Error dropping table '{table_name}': {e}",
                )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def _get_milvus_operator(self, operator: FilterOperator) -> str:
        mapping = {
            FilterOperator.EQUAL: "==",
            FilterOperator.NOT_EQUAL: "!=",
            FilterOperator.LESS_THAN: "<",
            FilterOperator.LESS_THAN_OR_EQUAL: "<=",
            FilterOperator.GREATER_THAN: ">",
            FilterOperator.GREATER_THAN_OR_EQUAL: ">=",
            FilterOperator.IN: "in",
            FilterOperator.NOT_IN: "not in",
            FilterOperator.LIKE: "like",
            FilterOperator.NOT_LIKE: "not like",
        }
        if operator not in mapping:
            raise Exception(f"Operator {operator} is not supported by Milvus!")
        return mapping[operator]

    def _translate_conditions(self, conditions: Optional[List[FilterCondition]], exclude_id: bool = True) -> Optional[str]:
        """
        Translate a list of FilterCondition objects a string that can be used by Milvus.
        E.g.,
        [
            FilterCondition(
                column="metadata.price",
                op=FilterOperator.LESS_THAN,
                value=1000,
            ),
            FilterCondition(
                column="metadata.price",
                op=FilterOperator.GREATER_THAN,
                value=300,
            )
        ]
        Is converted to: "(price < 1000) and (price > 300)"
        If exclude_id is set to true then id column is ignored
        """
        if not conditions:
            return
        # Ignore all non-metadata conditions
        filtered_conditions = [
            condition
            for condition in conditions
            if condition.column.startswith(TableField.METADATA.value) or condition.column.startswith(TableField.ID.value)
        ]
        if len(filtered_conditions) == 0:
            return None
        # Translate each metadata condition into a dict
        milvus_conditions = []
        for condition in filtered_conditions:
            if isinstance(condition.value, str):
                condition.value = f"'{condition.value}'"
            milvus_conditions.append(f"({condition.column.split('.')[-1]} {self._get_milvus_operator(condition.op)} {condition.value})")
        # Combine all metadata conditions into a single string and return
        return " and ".join(milvus_conditions) if milvus_conditions else None

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> HandlerResponse:
        # Load collection table
        collection = Collection(table_name)
        try:
            collection.load()
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Error loading collection {table_name}: {e}",
            )

        # Find vector filter in conditions
        vector_filter = (
            []
            if conditions is None
            else [
                condition.value
                for condition in conditions
                if condition.column == TableField.SEARCH_VECTOR.value
            ]
        )

        # Generate search arguments
        search_arguments = {
            "param": {}
        }
        # TODO: check if distance in columns work
        if columns:
            search_arguments["output_fields"] = columns
        else:
            search_arguments["output_fields"] = [schema_obj.name for schema_obj in self.SCHEMA]
        search_arguments["expr"] = self._translate_conditions(conditions)
        # NOTE: According to api sum of offset and limit should be less than 16384.
        api_limit = 16384
        if limit is not None and offset is not None and limit + offset >= api_limit:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Sum of limit and offset should be less than {api_limit}",
            )
        if limit is not None:
            search_arguments["limit"] = limit
        else:
            search_arguments["limit"] = self._search_limit
        if offset is not None:
            search_arguments["param"]["offset"] = offset

        # Vector search
        if vector_filter:
            search_arguments["data"] = vector_filter
            search_arguments["anns_field"] = TableField.EMBEDDINGS.value
            search_arguments["param"]["metric_type"] = self._search_params["search_metric_type"]
            search_arguments["param"]["ignore_growing"] = self._search_params["search_ignore_growing"]
            search_arguments["param"]["param"] = self._search_params["search_ignore_growing"]
            results = collection.search(**search_arguments)
            columns_required = [TableField.ID.value, TableField.DISTANCE.value]
            if TableField.CONTENT.value in columns:
                columns_required.append(TableField.CONTENT.value)
            if TableField.EMBEDDINGS.value in columns:
                columns_required.append(TableField.EMBEDDINGS.value)
            data = {k: [] for k in columns_required}
            for hits in results:
                for hit in hits:
                    for col in columns_required:
                        if col != TableField.DISTANCE.value:
                            data[col].append(hit.entity.get(col))
                        else:
                            data[TableField.DISTANCE.value].append(hit.distance)
            return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=pd.DataFrame(data))
        else:
            # Basic search
            if not search_arguments["expr"]:
                search_arguments["expr"] = ""
            search_arguments["output_fields"] = [
                TableField.ID.value,
                TableField.CONTENT.value,
                TableField.EMBEDDINGS.value,
            ] if not columns else columns
            results = collection.query(**search_arguments)
            return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=pd.DataFrame.from_records(results))

    def create_table(self, table_name: str, if_not_exists=True) -> HandlerResponse:
        """Create a collection with default parameters in the Milvus database as described in documentation."""
        id = FieldSchema(
            name=TableField.ID.value,
            dtype=DataType.VARCHAR,
            is_primary=True,
            max_length=self._create_table_params["create_id_max_len"],
            auto_id=self._create_table_params["create_auto_id"]
        )
        embeddings = FieldSchema(
            name=TableField.EMBEDDINGS.value,
            dtype=DataType.FLOAT_VECTOR,
            dim=self._create_table_params["create_embedding_dim"]
        )
        content = FieldSchema(
            name=TableField.CONTENT.value,
            dtype=DataType.VARCHAR,
            max_length=self._create_table_params["create_content_max_len"],
            default_value=self._create_table_params["create_content_default_value"]
        )
        schema = CollectionSchema(
            fields=[id, content, embeddings],
            description=self._create_table_params["create_schema_description"],
            enable_dynamic_field=self._create_table_params["create_dynamic_field"]
        )
        collection_name = table_name
        collection = None
        try:
            collection = Collection(
                name=collection_name,
                schema=schema,
                using=self._create_table_params["create_alias"],
            )
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Unable to create collection `{table_name}`: {e}"
            )
        try:
            collection.create_index(
                field_name=TableField.EMBEDDINGS.value,
                index_params={
                    "index_type": self._create_table_params["create_index_type"],
                    "metric_type": self._create_table_params["create_index_metric_type"],
                    "params": self._create_table_params["create_index_params"]
                }
            )
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Unable to create index on collection `{table_name}`: {e}"
            )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def insert(
        self, table_name: str, data: pd.DataFrame, columns: List[str] = None
    ) -> HandlerResponse:
        """Insert data into the Milvus collection."""
        collection = None
        try:
            collection = Collection(table_name)
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Unable to fetch collection `{table_name}`: {e}"
            )
        try:
            data = data[columns]
            if TableField.METADATA.value in data.columns:
                rows = data[TableField.METADATA.value].to_list()
                data = pd.concat([data, pd.DataFrame.from_records(rows)], axis=1)
                data.drop(TableField.METADATA.value, axis=1, inplace=True)
            collection.insert(data.to_dict(orient="records"))
            collection.flush()
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Unable to insert data into collection `{table_name}`: {e}"
            )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def delete(
        self, table_name: str, conditions: List[FilterCondition] = None
    ) -> HandlerResponse:
        # delete only supports IN operator
        for condition in conditions:
            if condition.op in [FilterOperator.EQUAL, FilterOperator.IN]:
                condition.op = FilterOperator.IN
                if not isinstance(condition.value, list):
                    condition.value = [condition.value]
        filters = self._translate_conditions(conditions, exclude_id=False)
        if not filters:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message="Some filters are required, use DROP TABLE to delete everything",
            )
        try:
            collection = Collection(table_name)
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Error retrieving collection `{table_name}`: {e}",
            )
        try:
            collection.delete(filters)
            collection.flush()
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Error deleting from collection `{table_name}`: {e}",
            )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def get_columns(self, table_name: str) -> HandlerResponse:
        """Get columns in a Milvus collection"""
        collection = None
        try:
            collection = Collection(table_name)
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Error finding table: {e}",
            )
        try:
            field_names = {field["name"] for field in collection.schema.fields}
            schema = [mindsdb_schema_field for mindsdb_schema_field in self.SCHEMA if mindsdb_schema_field["name"] in field_names]
            data = pd.DataFrame(schema)
            data.columns = ["COLUMN_NAME", "DATA_TYPE"]
            return HandlerResponse(data_frame=data)
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Error finding table: {e}",
            )



from mindsdb_sql.parser.ast import (
    BinaryOperation,
    Identifier,
    Constant,
    Update,
)
from mindsdb_sql.planner.steps import UpdateToTable
from mindsdb_sql.planner.utils import query_traversal

from mindsdb.api.executor.sql_query.result_set import ResultSet
from mindsdb.api.executor.exceptions import WrongArgumentError

from .base import BaseStepCall


class UpdateToTableCall(BaseStepCall):

    bind = UpdateToTable

    def call(self, step):
        data = ResultSet()

        if len(step.table.parts) > 1:
            integration_name = step.table.parts[0]
            table_name_parts = step.table.parts[1:]
        else:
            integration_name = self.context['database']
            table_name_parts = step.table.parts

        dn = self.session.datahub.get(integration_name)

        result = step.dataframe

        params_map_index = []

        if step.update_command.keys is not None:
            result_data = result.result_data

            where = None
            update_columns = {}

            key_columns = [i.to_string() for i in step.update_command.keys]
            if len(key_columns) == 0:
                raise WrongArgumentError('No key columns in update statement')
            for col in result_data.columns:
                name = col.name
                value = Constant(None)

                if name in key_columns:
                    # put it to where

                    condition = BinaryOperation(
                        op='=',
                        args=[Identifier(name), value]
                    )
                    if where is None:
                        where = condition
                    else:
                        where = BinaryOperation(
                            op='and',
                            args=[where, condition]
                        )
                else:
                    # put to update
                    update_columns[name] = value

                params_map_index.append([name, value])

            if len(update_columns) is None:
                raise WrongArgumentError(f'No columns for update found in: {result_data.columns}')

            update_query = Update(
                table=Identifier(parts=table_name_parts),
                update_columns=update_columns,
                where=where
            )

        else:
            # make command
            update_query = Update(
                table=Identifier(parts=table_name_parts),
                update_columns=step.update_command.update_columns,
                where=step.update_command.where
            )

            if result is None:
                # run as is
                dn.query(query=update_query, session=self.session)
                return data
            result_data = result.result_data

            # link nodes with parameters for fast replacing with values
            input_table_alias = step.update_command.from_select_alias
            if input_table_alias is None:
                raise WrongArgumentError('Subselect in update requires alias')

            def prepare_map_index(node, is_table, **kwargs):
                if isinstance(node, Identifier) and not is_table:
                    # is input table field
                    if node.parts[0] == input_table_alias.parts[0]:
                        node2 = Constant(None)
                        param_name = node.parts[-1]
                        params_map_index.append([param_name, node2])
                        # replace node with constant
                        return node2
                    elif node.parts[0] == table_name_parts[0]:
                        # remove updated table alias
                        node.parts = node.parts[1:]

            # do mapping
            query_traversal(update_query, prepare_map_index)

        # check all params is input data:
        data_header = [col.alias for col in result_data.columns]

        for param_name, _ in params_map_index:
            if param_name not in data_header:
                raise WrongArgumentError(f'Field {param_name} not found in input data. Input fields: {data_header}')

        # perform update
        for row in result_data.get_records():
            # run update from every row from input data

            # fill params:
            for param_name, param in params_map_index:
                param.value = row[param_name]

            dn.query(query=update_query, session=self.session)
        return data
